#!/usr/bin/perl -w

#
# This is a SAS Component
#

#
# Copyright (c) 2003-2006 University of Chicago and Fellowship
# for Interpretations of Genomes. All Rights Reserved.
#
# This file is part of the SEED Toolkit.
#
# The SEED Toolkit is free software. You can redistribute
# it and/or modify it under the terms of the SEED Toolkit
# Public License.
#
# You should have received a copy of the SEED Toolkit Public License
# along with this program; if not write to the University of Chicago
# at info@ci.uchicago.edu or the Fellowship for Interpretation of
# Genomes at veronika@thefig.info or download a copy from
# http://www.theseed.org/LICENSE.TXT.
#

use strict;
use Data::Dumper;
use Carp;
use Getopt::Long;
Getopt::Long::Configure("pass_through");

=head1 svr_align_seqs

    svr_align_seqs [options] < seqs.fa > ali.fa

This script takes a FASTA file from the standard input and writes the
alignment in the FASTA format to the standard output. When not
possible, a message will be written to the standard error output.

Examples:
   
  Use MUSCLE to align sequences in seqs.fa:

    svr_align_seqs -tool=muscle < seqs.fa > ali.fa

  Use MAFFT to do profile alignment of seqs.fa with seed alignment in profile.fa:
   
    svr_align_seqs -tool=mafft -profile=profile.fa < seqs.fa > ali.fa

=head2 Command-Line Options

=over 4

=head3 Common options

=over 4

=item tool

Alignment program to use. Supported tools: muscle, mafft and clustal (default). 

=item profile

FASTA file name of skeleton alignment to be aligned with input
sequences. Supported in MUSCLE and MAFFT.

=item version

Print tool version information. Supported in MUSCLE and MAFFT.

=back

=head3 MUSCLE options

  Supported values: anchorspacing center cluster1 cluster2 diagbreak diaglength diagmargin distance1 distance2 gapopen log loga matrix maxhours maxiters maxmb maxtrees minbestcolscore minsmoothscore objscore refinewindow root1 root2 scorefile seqtype smoothscorecell smoothwindow spscore SUEFF usetree weight1 weight2

  Supported flags: anchors brenner cluster dimer diags diags1 diags2 le noanchors quiet sp spn stable sv verbose

=head3 MAFFT options

  Algorithm aliases: -alg (linsi | einsi | ginsi | nwnsi | nwns | fftnsi | fftns (D)) (in descending order or accuracy)

  Supported values: aamatrix bl ep groupsize jtt lap lep lepx LOP LEXP maxiterate op partsize retree tm weighti

  Supported flags: 6merpair amino auto clustalout dpparttree fastapair fastaparttree fft fmodel genafpair globalpair inputorder localpair memsave nofft noscore nuc parttree quiet reorder treeout

=head2 Output Format

The standard output is the alignment in the FASTA format.

=cut

use gjoalignment;
use gjoseqlib;

my $usage = [ "Usage: $0 [options] < seqs.fa > ali.fa","",
              "\tOPTIONAL","",
              "\t  -tool     clustal (D), muscle, mafft",
              "\t  Common options supported in muscle and mafft:",
              "\t  -profile  FASTA file name of skeleton alignment to be aligned with input sequences",
              "\t  -version  Print tool version information", "",
              "\tTool-specific options", "",
              "\t  Muscle options:",
              "\t    Values: anchorspacing center cluster1 cluster2 diagbreak diaglength diagmargin",
              "\t            distance1 distance2 gapopen log loga matrix maxhours maxiters maxmb",
              "\t            maxtrees minbestcolscore minsmoothscore objscore refinewindow root1",
              "\t            root2 scorefile seqtype smoothscorecell smoothwindow spscore SUEFF",
              "\t            usetree weight1 weight2",
              "\t    Flags:  anchors brenner cluster dimer diags diags1 diags2 le noanchors quiet",
              "\t            sp spn stable sv verbose", "",
              "\t  Mafft options:",
              "\t    Algorithm aliases: -alg (linsi | einsi | ginsi | nwnsi | nwns | fftnsi | fftns (D))",
              "\t            (in descending order or accuracy)",
              "\t    Values: aamatrix bl ep groupsize jtt lap lep lepx LOP LEXP maxiterate",
              "\t            op partsize retree tm weighti",
              "\t    Flags:  6merpair amino auto clustalout dpparttree fastapair fastaparttree",
              "\t            fft fmodel genafpair globalpair inputorder localpair memsave nofft",
              "\t            noscore nuc parttree quiet reorder treeout", "", "" ];

my $help;
my $version;
my $profileF; 
my $tool = "clustal";
my ($clustal, $mafft, $muscle);

my $opted = GetOptions("help"           => \$help,
                       "version"        => \$version,
                       "tool|program=s" => \$tool,
                       "clustal"        => \$clustal,
                       "mafft"          => \$mafft,
                       "muscle"         => \$muscle,
                       "profile=s"      => \$profileF,
                       '<>'             => \&remaining_options
                      );

$help and die join("\n", @$usage);

my $opts;    

if ($profileF) {
    my $profile = gjoseqlib::read_fasta($profileF);
    $opts->{profile} = $profile;
}


my %mafft_val = map { $_ => 1 }
                qw( alg algorithm

                    aamatrix
                    bl
                    ep
                    groupsize
                    jtt
                    lap
                    lep
                    lepx
                    LOP
                    LEXP
                    maxiterate
                    op
                    partsize
                    retree
                    tm
                    weighti
                 );

my %mafft_flag = map { $_ => 1 }
                 qw( 6merpair
                     amino
                     auto
                     clustalout
                     dpparttree
                     fastapair
                     fastaparttree
                     fft
                     fmodel
                     genafpair
                     globalpair
                     inputorder
                     localpair
                     memsave
                     nofft
                     noscore
                     nuc
                     parttree
                     quiet
                     reorder
                     treeout
                  );

my %muscle_val  = map { $_ => 1 }
                  qw( anchorspacing
                      center
                      cluster1
                      cluster2
                      diagbreak
                      diaglength
                      diagmargin
                      distance1
                      distance2
                      gapopen
                      log
                      loga
                      matrix
                      maxhours
                      maxiters
                      maxmb
                      maxtrees
                      minbestcolscore
                      minsmoothscore
                      objscore
                      refinewindow
                      root1
                      root2
                      scorefile
                      seqtype
                      smoothscorecell
                      smoothwindow
                      spscore
                      SUEFF
                      usetree
                      weight1
                      weight2
                   );

my %muscle_flag = map { $_ => 1 }
                  qw( anchors
                      brenner
                      cluster
                      dimer
                      diags
                      diags1
                      diags2
                      le
                      noanchors
                      quiet
                      sp
                      spn
                      stable
                      sv
                      verbose
                   );

my $program;

my @prog_args;;

my $prog_vals;
my $prog_flags;

if    ($tool =~ /mafft/i  || $mafft)  { $program = \&align_with_mafft;   $prog_vals = \%mafft_val;  $prog_flags = \%mafft_flag;  }
elsif ($tool =~ /muscle/i || $muscle) { $program = \&align_with_muscle;  $prog_vals = \%muscle_val; $prog_flags = \%muscle_flag; }
else                                  { $program = \&align_with_clustal; }


if ($version) {
    $version = $program->(1, { version => 1 });
    print "$version\n";
} else {
    my $opts = process_prog_args(\@prog_args);
    
    my $seqs = gjoseqlib::read_fasta();
    my $ali = $program->($seqs, $opts);

    gjoseqlib::print_alignment_as_fasta($ali);
}

sub remaining_options {
    push @prog_args, $_[0];
}

sub process_prog_args {
    my @args = @{$_[0]};
    my %opts;
    while ((shift @args) =~ /^-(\S+)/) {
        my $arg = $1;
        if (defined($prog_vals) && defined($prog_flags)) {
            if ($prog_vals->{$arg}) {
                $opts{$arg} = shift @args;
            } elsif ($prog_flags->{$arg}) {
                $opts{$arg} = 1;
            } else {
                die "Bad flag: '$arg'\n\n" . join("\n", @$usage)."\n";
            }
        } else {
            if (@args > 0 && $args[0] !~ /^-/) {
                $opts{$arg} = shift @args;
            } else {
                $opts{$arg} = 1;
            }
        }
    }
    return \%opts;
}
